# Global variable used for model fitting
fit_model_data = NULL    

# Fits segmented model n_max times with shifting initial breaks starting at guess_breaks;
# for a 3-break model, breaks slide from (a,b,c) to (a-max_width, b, c+max_width);
# for a 2-break model, breaks slide from (a,c) to (a-max_width, c+max_width);
# for a 1-break model, break slides from (c) to (c+max_width);
fit_canyon_iteratively = function(lin_model, guess_breaks, sunrise, max_width=2, n_max=30) {
  # Function that checks if the model was successfull
  no_errors = function(model) all(class(model)!="try-error")
  # Set up breaks
  k = max_width/n_max
  breaks_incr = list(k, k*c(-1,1), k*c(-1, 0, 1))
  breaks_incr = breaks_incr[[length(guess_breaks)]]
  cur_breaks = guess_breaks
  # List to keep all models
  models = list()
  # Fit model iteratively
  for (i in 1:n_max) {
    model = try(segmented(lin_model, seg.Z=~Hour, psi=list(Hour=cur_breaks)), silent=TRUE)
    if (no_errors(model)) models = c(models, list(model))
    cur_breaks = cur_breaks + breaks_incr
  }
  # Select best model (highest r2)
  best_model = NULL
  if (length(models)>0) {
    r2 = laply(models, function(x) summary(x)$r.squared)
    ix = which(r2==max(r2))
    best_model = models[unique(ix)][[1]]
  } 
  # Search for a canyon model some more
  if (is.null(best_model) & n_max<200) {
    fit_canyon_iteratively(lin_model, guess_breaks, sunrise, max_width, 2*n_max)
  } else {
  # Return best model and simple statistics
    list(BestModel=best_model, FitRatio=length(models)/n_max)
  }
}

# Fits segmented model for one hive and date;
# includes data before final_hour()
fit_canyon = function(M) {
  U = unique_hive_date(M)
  print(paste("fit", model_number, "/" ,number_of_models, U$Hive, U$Date))
  assign("model_number", model_number + 1, envir=globalenv()) # A long way to go for a simple increment (!)
  # Select data before final hour
  final_hr = final_hour(U$Hive, U$Date)
  if (is.na(final_hr)) return(NULL)
  ix = with(M, to_hours(DateTime) < final_hr)
  M2 = M[ix,]
  if (nrow(M2) < 30) return(NULL)
  # Put data for regression in global variable
  fit_model_data <<- M2
  # Fit linear model as a basis for segmented regression
  lin_model= lm(Weight~Hour, data=fit_model_data)
  # Initial guesses for breaks are relative to sunrise
  sunrise = get_sunrise(U$Hive, U$Date)
  # Fit 3-break model
  model = fit_canyon_iteratively(lin_model, c(sunrise+0.5,sunrise+2,sunrise+3), sunrise) 
  # Or else, fit 2-break model 
  if (is.null(model$BestModel))
    model = fit_canyon_iteratively(lin_model, c(sunrise+0.5,sunrise+2), sunrise) 
  # Or else, fit 1-break model 
  if (is.null(model$BestModel))
    model = fit_canyon_iteratively(lin_model, sunrise+0.5, sunrise)
  # Return the model
  c(Hive=list(U$Hive), Date=list(U$Date), Model=model)
}

# Extracts segments from a segmented model;
# returns the number of segments and three data frames (or NULL):
# - breaks with SE
# - slopes with SE and intercepts
# - (x,y) data to plot each line
# Note that the segments do not connect event hough gap has been said false;
# this is corrected through a calculation of the displacement
extract_segments = function(model) {
  if (is.null(model)) return(NULL)
  breaks = get_breaks(model)
  slopes = get_slopes(model)
  # intercepts = intercept(model, gap=FALSE)[[1]][,1]
  intercepts = intercept(model)[[1]][,1]
  x = model$model$Hour
  pred = predict(model)
  endings = c(min(x), breaks, max(x))
  n = length(slopes)
  line_plot = data.frame()
  # Loop through segments
  for (i in 1:n) {
    a = slopes[i]
    b = intercepts[i]
    x1 = endings[i]
    x2 = endings[i+1]
    y1 = a*x1+b
    y2 = a*x2+b
    ix_pred = min(which(x>x1))
    displacement = a*x[ix_pred] + b - pred[ix_pred]
    line_plot = rbind(line_plot, data.frame(Line=paste0("L",i), X=x1, Y=y1-displacement))
    line_plot = rbind(line_plot, data.frame(Line=paste0("L",i), X=x2, Y=y2-displacement))
  }
  rownames(line_plot) = NULL
  list(
    NumSegments = n,
    Breaks = data.frame(Value=breaks, 
               SE=model$psi[,3],
               row.names=1:(n-1)),
    Lines = data.frame(Slope=slopes,
               SlopeSE=slope(model)[[1]][,2],
               Intercept=intercepts,
               row.names=1:n),
    LinePlot = line_plot
  )
}
  
# Extracts canyon information from slps[ix];
# returns a single-row data frame
extract_category = function(slps, ix) {
  # Last segment cannot be chosen
  if (nrow(slps) == ix) return(na_category) 
  data.frame(              
    CanyonSegment = ix,
    CanyonHour = slps$To[ix],
    CanyonHourBegin = slps$From[ix],
    CanyonHourEnd = slps$To[ix+1],
    CanyonSlope = slps$Slope[ix],
    CanyonSlopeUp = slps$Slope[ix+1],
    CanyonDepth = -slps$Slope[ix]*(slps$To[ix]-slps$From[ix])
  )                   
}

# Transforms a category code, retaining only those digits that refer to a segment starting before sunrise;
# returns the new (shortened) category code
night_category = function(slps, category) {
  has = function(digit) grepl(as.character(digit), as.character(category)) 
  sunrise = sunshine(slps)[1]
  new_cat = 0
  for (i in 1:4) {
    if (has(i)) if (slps$From[i] < sunrise) new_cat = 10*new_cat + i
  }
  new_cat
}

# Finds the night segment, i,e, the segment of the longest duration before sunrise;
# returns a data frame with info on that segment
extract_night = function(slps, category) {
  sunrise = sunshine(slps)[1]
  slps = mutate(slps, Duration=To-From)
  slps = slps[order(slps$Duration,decreasing=TRUE),]
  rim = category$CanyonHourBegin

  M = data.frame(
        NightFrom  = NA,
        NightTo    = NA,
        NightSlope = NA
      )
  if (is.na(rim)) return(M)
  
  ix = 1
  while (slps$To[ix]>rim) ix=ix+1
  data.frame(
    NightFrom = slps$From[ix],
    NightTo = slps$To[ix],
    NightSlope = slps$Slope[ix]
  )
}

# Returns an NA data frame
na_category = function() {
  data.frame(
    CanyonSegment   = NA,
    CanyonHour      = NA,
    CanyonHourBegin = NA,
    CanyonHourEnd   = NA,
    CanyonSlope     = NA,
    CanyonSlopeUp   = NA,
    CanyonDepth     = NA
  )                   
}

# Categorizes the segment slopes for a hive and date;
# returns a data frame with one row contain columns with information on canyon and night
categorize = function(slps) {
    # Which slopes are negative?
    neg = which(slps$Slope<0)
    # Turn vector into an integer number with as many digits that there are negative slopes;
    # each digit means the corresponding slope is negatuve
    course = if (length(neg)>0) as.numeric(paste(neg,collapse="")) else 0
    # Branch out according to the course;
    # cat will become a data frame with one row containing canyon information (segment number, hour, slope, depth)
    cat = if (course==0) {
      na_category()
    } else if (course==1) {
      na_category()
    } else if (course==2) {
      extract_category(slps, 2)
    } else if (course==3) {
      extract_category(slps, 3)
    } else if (course==4) {
      na_category()
    } else if (course==12) {
      extract_category(slps, 2)
    } else if (course==13) {
      extract_category(slps, 3)
    } else if (course==14) {
      na_category()
    } else if (course==23) {
      extract_category(slps, 3)
    } else if (course==24) {
      extract_category(slps, 2)
    } else if (course==34) {
      na_category()
    } else if (course==123) {
      extract_category(slps, 3)
    } else if (course==124) {
      extract_category(slps, 2)
    } else if (course==134) {
      na_category()
    } else if (course==234) {
      na_category()
    } else if (course==1234) {
      na_category()
    } else {
      stop(paste("Unknown course:", course))
    }
    cat = mutate(cat, Category = course, NumSegments = unique(slps$NumBreaks) + 1)
    cbind(cat, extract_night(slps, cat))
}

segment_models_overview = function() {
  M = ldply(slr_models, function(x) data.frame(FitRatio=x$Model.FitRatio, CanyonApparancy=x$Model.CanyonApparancy))
  M[,2:3]
}

find_segment_model = function(hive, date_) {
  ix_hive = which(sapply(slr_models, function(x) x$Hive==hive))
  ix_date = which(sapply(slr_models, function(x) x$Date==date_))
  ix = intersect(ix_hive, ix_date)
  slr_models[[ix]]
}
# find_segment_model("hive73", dmy("1/7/2014"))


#
# Main
#

load("output/W.Rdata")
load("output/WD.Rdata")
print("A")
# Set up counters
M = ddply(W, .(Hive,Date), summarize, Count=1)
number_of_models = nrow(M)
model_number = 1

# Do fit
slr_models = dlply(W, .(Hive,Date), fit_canyon)
slr_models = remove_null(slr_models)

# Keep only SLR models with 3 breaks (i.e., with 4 segments)
slr_models = llply(slr_models, function(x) if (nrow(x$Model.BestModel$psi) == 3) x else NULL)
slr_models = remove_null(slr_models)

save(slr_models, file="output/slr_models.Rdata")
