# Extracts canyon informations from slps[ix];
# returns a single-row data frame
extract_category = function(slps, ix) {
  # Last segment cannot be chosen
  if (nrow(slps) == ix) return(na_category) 
  M = data.frame(              
    CanyonSegment = ix,
    CanyonHour = slps$To[ix],         # Canyon bottom 
    CanyonHourBegin = slps$From[ix],  # Canyon left rim
    CanyonHourEnd = slps$To[ix+1],    # Canyon right rim
    CanyonSlope = slps$Slope[ix],
    CanyonSlopeUp = slps$Slope[ix+1],
    CanyonDepth = -slps$Slope[ix]*(slps$To[ix]-slps$From[ix]),
    Joined_2_3 = FALSE
  )   
  # Joined segment 2 and 3, if slope2 < slope3
  if (ix==3 & slps$Slope[2] < slps$Slope[3]) {
    M2 = data.frame(              
      CanyonSegment = 2,
      CanyonHour = M$CanyonHour,
      CanyonHourBegin = slps$From[2],
      CanyonHourEnd = M$CanyonHourEnd,
      CanyonSlopeUp = M$CanyonSlopeUp,
      CanyonDepth = -slps$Slope[2]*(slps$To[2]-slps$From[2]) + M$CanyonDepth,
      Joined_2_3 = TRUE
    )   
    M2$CanyonSlope = with(M2, -CanyonDepth/(CanyonHour - CanyonHourBegin))
    M = M2
  }
  M
}

# Transforms a category code, retaining only those digits that refer to a segment starting before sunrise;
# returns the new (shortened) category code
night_category = function(slps, category) {
  has = function(digit) grepl(as.character(digit), as.character(category)) 
  sunrise = sunshine(slps)[1]
  new_cat = 0
  for (i in 1:4) {
    if (has(i)) if (slps$From[i] < sunrise) new_cat = 10*new_cat + i
  }
  new_cat
}

# Finds the night segment, i,e, the segment of the longest duration before canyon rim;
# returns a data frame with info on that segment
extract_night = function(slps, category) {
  # Put slopes in order of decreasing slope
  slps = mutate(slps, Duration=To-From)
  slps = slps[order(slps$Duration,decreasing=TRUE),]
  
  # Locate canyon rim
  rim = category$CanyonHourBegin
  M = data.frame(
        NightFrom  = NA,
        NightTo    = NA,
        NightSlope = NA
      )
  if (is.na(rim)) return(M)
  
  # Pick the first (and longest segment ending before the rim)
  ix = 1
  while (slps$To[ix]>rim) ix=ix+1
  data.frame(
    NightFrom = slps$From[ix],
    NightTo = slps$To[ix],
    NightSlope = slps$Slope[ix]
  )
}

# Returns an NA data frame
na_category = function() {
  data.frame(
    CanyonSegment   = NA,
    CanyonHour      = NA,
    CanyonHourBegin = NA,
    CanyonHourEnd   = NA,
    CanyonSlope     = NA,
    CanyonSlopeUp   = NA,
    CanyonDepth     = NA
  )                   
}

# Categorizes the segment slopes for a hive and date;
# returns a data frame with one row contain columns with information on canyon and night
categorize = function(slps) {
    # print(paste(slps$Hive, slps$Date))
    # Which slopes are negative?
    neg = which(slps$Slope<0)
    # Turn vector into an integer number with as many digits that there are negative slopes;
    # each digit means the corresponding slope is negatuve
    course = if (length(neg)>0) as.numeric(paste(neg,collapse="")) else 0
    # Branch out according to the course;
    # cat will become a data frame with one row containing canyon information (segment number, hour, slope, depth)
    cat = if (course==0) {
      na_category()
    } else if (course==1) {
      na_category()
    } else if (course==2) {
      extract_category(slps, 2)
    } else if (course==3) {
      extract_category(slps, 3)
    } else if (course==4) {
      na_category()
    } else if (course==12) {
      extract_category(slps, 2)
    } else if (course==13) {
      extract_category(slps, 3)
    } else if (course==14) {
      na_category()
    } else if (course==23) {
      extract_category(slps, 3)
    } else if (course==24) {
      extract_category(slps, 2)
    } else if (course==34) {
      na_category()
    } else if (course==123) {
      extract_category(slps, 3)
    } else if (course==124) {
      extract_category(slps, 2)
    } else if (course==134) {
      na_category()
    } else if (course==234) {
      na_category()
    } else if (course==1234) {
      na_category()
    } else {
      stop(paste("Unknown course:", course))
    }
    cat = mutate(cat, Category = course, NumSegments = unique(slps$NumBreaks) + 1, CanyonR2=unique(slps$R2))
    cbind(cat, extract_night(slps, cat))
}

#
# Main
#

# Load weight data and best SLR models
load("output/WD.Rdata")
load("output/best_slr.Rdata")

# Collate Breakfast Canyon information
canyons = ddply(best_slr, .(Hive,Date), categorize)
canyons$Category = factor(canyons$Category)
canyons$CanyonSegment = factor(canyons$CanyonSegment)
canyons$NightDuration = canyons$NightTo - canyons$NightFrom
canyons = join(canyons, WD[,c("Hive","Date","Sunrise")])
canyons = mutate(canyons, CanyonDayHour=CanyonHour-Sunrise)
canyons = join(canyons, WD[,c("Hive","Date","WeightGain")])
canyons = mutate(canyons, WeightGainHr=WeightGain/24)

# Discard outliers
outliers = find_outliers(canyons$NightSlope)
n = nrow(subset(outliers, IsOutlier))
paste("Number of outliers discarded", n)
canyons = canyons[!outliers$IsOutlier,]
nrow(canyons)

# Save
save(canyons, file="output/canyons.Rdata")
  