# Define when pesticide treatments started
treatment_dates = data.frame(Year=c(2014,2015), TreatmentDate=dmy("15/7/2014", "10/7/2015"))

# Returns list with NULL values removed
remove_null = function(lst) {
  ix = which(sapply(lst, is.null))
  if (length(ix)==0) lst else lst[-ix]
}

# Differences between values in vector headed by an NA
na_diff = function(x) {
  c(NA, diff(x))
}

# Lag 2-differences between values in vector headed by an 2 NAs
na_diff2 = function(x) {
  c(NA, NA, diff(x, lag=2))
}

# One-hive subset of data frame 
subset_hive = function(M, hive) {
  subset(M, Hive==hive)
}

# Convert string to date
to_date = function(s) {
  parse_date_time(s, c("dmy", "ymd"))
}

# Convert string to date-time
to_date_time = function(s) {
  parse_date_time(s, c("dmyHMS", "ymdHMS"))
}

# Subset of data frame between dates
subset_date = function(M, from, to=from) {
  subset(M, Date>=to_date(from) & Date<=to_date(to))
}

# One-hive subset of data frame between dates 
subset_hive_date = function(M, hive, from, to=from) {
  subset(M, Hive==hive & Date>=to_date(from) & Date<=to_date(to))
}

# Row indices for one hive between date-times, including one date-time before and after
which_hive_date_time = function(M, hive, from, to=from) {
  which(M$Hive==hive & M$DateTime>=to_date_time(from)-1 & M$DateTime<=to_date_time(to)+1)
}

# Check that data frame contains only one hive and return the hive name
unique_hive = function(M) {
  hive = unique(M$Hive)
  assert_that(length(hive) == 1)
  hive
}

# Check that data frame contains only one date hive and return the date
unique_date = function(M) {
  date = unique(M$Date)
  assert_that(length(date) == 1)
  date
}

# Check that data frame contains only one hive and one date 
# and return hive name and date in a list
unique_hive_date = function(M) {
  hive = unique(M$Hive)
  date = unique(M$Date)
  assert_that(length(hive) == 1)
  assert_that(length(date) == 1)
  list(Hive=hive,Date=date)
}

# Takes a vector that contains runs of equal dates;
# returns a vector which marks the first and last index of every run of equal dates
mark_ends = function(x) {
  ix = which(c(FALSE, diff(day(x))!=0))
  ix_first = ix[1]
  ix_last = tail(ix,1) - 1
  at_end = rep(TRUE, length(x))
  at_end[ix_first:ix_last] = FALSE
  at_end
}

# Computes hours [0;24] from date-time
to_hours = function(date_time) {
  as.numeric(difftime(date_time, date(date_time), units="hours"))
}

# Computes sunrise and sunset for a data frame that contains one date and one hive
sun = function(M, latitude, longitude, time_zone) {
  U = unique_hive_date(M)
  noon = as.POSIXct(U$Date) + 12*3600
  T = solar_time(latitude, longitude, time_zone, noon)
  data.frame(Sunrise=to_hours(T$SunriseSolarTime), Sunset=to_hours(T$SunsetSolarTime))
}

# Returns sunrise and sunset for one date and one hive
sunshine = function(M) {
  U = unique_hive_date(M)
  today = subset_hive_date(WD, as.character(U$Hive), as.character(U$Date))
  c(sunrise = today$Sunrise,  sunset = today$Sunset)
}

# Returns sunrise for a certain hive and date
get_sunrise = function(hive, date) {
  subset_hive_date(WD, hive, date)$Sunrise
}

# Returns equilibrium moisture content for honey at temperature and %RH
emc = function (T,RH) {
  h=RH/100
  W = 349 + 1.29*T + 0.0135*T^2
  K = 0.805 + 0.000736*T - 0.00000273*T^2
  K1 = 6.27 - 0.00938*T - 0.000303*T^2
  K2 = 1.91 + 0.0407*T - 0.000293*T^2
  Kh = K*h
  a = K1*Kh
  b = K1*K2*K^2*h^2
  1800/W*( Kh/(1-Kh) + (a+2*b)/(1+a+b) )
}

# Takes same-length vectors of weight and weight change rate (kg/h);
# returns a same-length vector with values that should be dropped (TRUE) or not (FALSE);
# will detect singular or double outliers
noise_drop = function(weight, weight_rate) {
  n = length(weight)
  drop = rep(FALSE, n)
  # 3-points
  i = 1
  while (i<n-2) {
    # print(paste(i,n))
    r = weight_rate[i:(i+2)]
    w = weight[i:(i+2)]
    dw13 = abs(w[1] - w[3])
    dw12 = abs(w[1] - w[2])
    if (!any(is.na(c(r,w))) & abs(r[2]) > 1.2  &  dw13 < 0.25*dw12 ) {
      drop[i+1] = TRUE
      i = i+2
    } else {
      i = i+1
    }
  }
  # 4-points
  i = 1
  while (i<n-3) {
    # print(paste(i,n))
    r = weight_rate[i:(i+3)]
    w = weight[i:(i+3)]
    dw14 = abs(w[1] - w[4])
    dw12 = max(abs(w[1] - w[2]), abs(w[1] - w[3]))
    if (!any(is.na(c(r,w))) & (abs(r[2]) > 1.2 | abs(r[3]) > 1.2) & dw14 < 0.20*dw12 ) {
      drop[i+1] = TRUE
      drop[i+2] = TRUE
      i = i+3
    } else {
      i = i+1
    }
  }
  drop
}
      
# Read weather file, replacing time with solar time      
read_weather = function(file_name, latitude, longitude, time_zone, time_format) {
  # Read file
  M = read.table(file=file_name, header=TRUE, sep="\t", stringsAsFactors=FALSE  ) 
  M$DateTime = parse_date_time(M$DateTime, time_format)
  # Change to solar date time
  T = solar_time(latitude, longitude, time_zone, M$DateTime)
  M$DateTime=T$SolarDateTime
  # Add date and hour
  M = mutate(M, Date = date(DateTime),
                Hour = as.numeric(difftime(DateTime, Date, units="hours")))
  # Add moisture content
  M$Emc = emc(M$Tavg, M$RH)
  # Move rain fall to beginning of recording interval
  n = nrow(M)
  M$Rain = c(M$Rain[2:n], NA)
  M
}   

# Returns a list of two data frames: obs on original time scale (W) and obs summarized daily (WD)
read_obs = function(inputs) {
  # Read weather file
  weather <<- read_weather(inputs$weather_file, inputs$latitude, inputs$longitude, inputs$time_zone, inputs$weather_file_date)
  # Read weight file
  M = read.table(file=inputs$weight_file, header=TRUE, sep="\t", stringsAsFactors=FALSE) 
  # Parse date-time
  ix_date = which(colnames(M)=="date" | colnames(M)=="Date")
  if (length(ix_date) != 1) stop("Weather files must have a column named 'date' or 'Date'")
  M$DateTime = parse_date_time(M[,ix_date], inputs$weight_file_date)
  M[,ix_date] = NULL
  # Create weight column
  M = melt(M, id.vars="DateTime", variable.name="Hive", value.name="Weight" )
  # Change to solar date time
  T = solar_time(inputs$latitude, inputs$longitude, inputs$time_zone, M$DateTime)
  M$DateTime=T$SolarDateTime
  # Add year, date and hour
  M = mutate(M, Year = year(DateTime),
                Date = date(DateTime),
                Hour = as.numeric(difftime(DateTime, Date, units="hours")))
  # Discard missing obs
  M = M[complete.cases(M),]
  # Add rain
  f = approxfun(weather$DateTime, weather$Rain, method ="constant")
  M$Rain = f(M$DateTime) 
  # Mark ends
  M = ddply(M, .(Hive), transform, AtEnd = mark_ends(DateTime))
  # Add time and weight steps      
  M = ddply(M, .(Hive), mutate, 
      WeightStep=na_diff(Weight),
      TimeStep=na_diff(DateTime)/60, 
      WeightRate=WeightStep/TimeStep
    )

  # Mark rows to drop due to weight noise
  M = ddply(M, .(Hive), transform,
        NoiseDrop = noise_drop(Weight, WeightRate)
      )
  # Show amount of weight noise dropped
  noise_drop = ddply(M, .(Year), summarise, TotalObs=length(NoiseDrop), NoiseDropObs=sum(NoiseDrop))
  print(noise_drop)
  # Drop weight noise
  M = subset(M, !NoiseDrop)
  M$NoiseDrop = NULL
  # Re-calculate time and weight steps      
  M = ddply(M, .(Hive), mutate, 
      WeightStep=na_diff(Weight),
      TimeStep=na_diff(DateTime)/60, 
      WeightRate=WeightStep/TimeStep
    )
  
  # Add row numbers
  rownames(M) = NULL
  M$Row = as.numeric(rownames(M))
 
  # Create daily values
  D = ddply(M, .(Hive, Date), function(X) sun(X, inputs$latitude, inputs$longitude, inputs$time_zone)  )
  D = join(D, 
            ddply(M, .(Hive,Date), summarize, 
              Rain = sum(Rain), 
              MaxWeightRate = max(abs(WeightRate))
            )
          )
          
  # Set location and year
  M$Location = factor(inputs$location)
  D$Location = factor(inputs$location)
  M$Year = factor(year(M$DateTime))
  D$Year = factor(year(D$Date))
  
  # Exclude dates
  D$Excluded = FALSE
  if (!is.null(inputs$exclude_dates)) {
    ix = D$Date %in% inputs$exclude_dates
    D$Excluded[ix] = TRUE
  }

  D = bc_add_treatments(D, c("input/treatments-az-2014.txt","input/treatments-az-2015.txt"))
  X = bc_add_statistics(M, D)

  # Cull ends
  M = cull(X[[1]])
  D = ddply(X[[2]], .(Hive), cull_daily)
  list(W=M, WD=D)
}

# Regresses a sine-cosine wave on same-length vectors of hour and detrended weight;
# Returns the lm model, or NULL if there are too few observations
regress_weight_model = function(hour, weight_detrended) {
  assert_that(length(hour) == length(weight_detrended))
  M = data.frame(y=weight_detrended,x=hour/24*2*pi)
  M = M[complete.cases(M),]
  if (nrow(M)>30) lm(y~cos(x) + sin(x), data=M) else NULL
}

# Returns the predicted values of a daily sine-cosine wave describing the detrended weight;
# takes same-length vectors of hour and detrended weight as input;
# returns the fitted values or a same-length vector of NA
regress_weight_curve = function(hour, weight_detrended) {
  model = regress_weight_model(hour, weight_detrended)
  if (is.null(model)) rep(NA, length(hour)) else model$fitted.values
}

# Returns amplitude and r2 for the detrended weight for a hive and date, or c(NA,NA)
regress_weight_daily = function(M) {
  unique_hive_date(M)
  model = regress_weight_model(M$Hour, M$WeightDetrended)
  amplitude = NA
  r2 = NA
  if (!is.null(model)) {
    amplitude = 2*sqrt(coef(model)[2]^2 + coef(model)[3]^2)
    r2 = summary(model)$r.squared
  }
  data.frame(Amplitude=amplitude, R2=r2)
}

# Removes the first and last observation as marked by AtEnd, i.e. the first and last observation for each hive
cull = function(M) {
  M = M[!M$AtEnd,]
  M$Row = NULL
  M$AtEnd = NULL
  M$WeightReset = NULL
  M$WeightCorrection = NULL
  rownames(M) = NULL
  M
}

# Removes the first two and the last observation
cull_daily = function(M) {
  n = nrow(M)
  M = M[-c(1,2, n),]
  rownames(M) = NULL
  M
}

# Estimates weight at midnight for one hive on one date, midnight considered as the beginning of that date;
# a parabolic curve is fitted on weight data from midnight plus/minus one hour;
# returns the weight and slope of weight change at midnight, or NA 
weight_at_midnight = function(M) {
  U = unique_hive_date(M)
  midnight = as.POSIXct(U$Date)
  from = midnight - 3600
  to = midnight + 3600
  n = 0
  repeat {
    w = subset(M, Hive==U$Hive & DateTime>from & DateTime<to)
    if (nrow(w)>3 | n==3) break;
    from = from - 3600
    to = to + 3600
    n = n + 1
  }
  
  value = NA
  slope = NA
  if (nrow(w)>3) {
    w = w[,c("Hour", "Weight")]
    w = w[complete.cases(w),]
    x = w$Hour
    x[x>12] = x[x>12] -24
    y = w$Weight
    model = lm(y ~ x+I(x^2), data=data.frame(x=x,y=y))
    # f(x)  = ax^2 +bx + c  =>  f(0) = c 
    value = coef(model)[1]
    #  f'(x) = 2ax + b        => f'(0) = b
    slope = coef(model)[2]
  }
  data.frame(WeightMidnight=value, SlopeMidnight=slope)
}

# Returns the weight at midnight at the end of the day for a hive and date
weight_at_end = function(hive, date) {
  tomorrow = subset_hive_date(WD, hive, date+1, date+1)
  value = NA
  if (nrow(tomorrow)==1) {
    w = tomorrow$WeightMidnight
    value = if (is.na(w)) NA else w
  }
  value
}

# Detrends the weight for a hive with one date by drawing a line between weight at the two midnights encompassing the day;
# returns a data frame with the original date times and the detrended weight.
detrend_weight = function(M, WD) {
  U = unique_hive_date(M)
  today = subset(WD, Hive==U$Hive & Date==U$Date)
  tomorrow = subset(WD, Hive==U$Hive & Date==U$Date+1)
  x = M$Hour
  y = rep(NA, length(x))
  if (nrow(today)==1 & nrow(tomorrow)==1) {
    y0 = today$WeightMidnight
    y1 = tomorrow$WeightMidnight
    x0 = 0
    x1 = 24
    w = y0 + (x-x0)/(x1-x0)*(y1-y0)
    y = M$Weight - w
  }
  data.frame(DateTime=M$DateTime, WeightDetrended = y)
}

# Reads the list of inputs;
# returns two data frames in a list: one on the original time resolution and one with daily values
bc_load = function(inputs) {
  f = function(x, name) x[name][[1]] 
  A = llply(inputs, read_obs)
  list(
    W  = ldply(A, f, name="W"),
    WD = ldply(A, f, name="WD")
  )  
}

# Add summary statistics to W and WD data frames
bc_add_statistics = function(M, D) {
  D = join(D, ddply(M, .(Hive,Date), weight_at_midnight))
  M  = join(M, ddply(M, .(Hive,Date), detrend_weight, WD=D))
  M  = ddply(M, .(Hive,Date), mutate, WeightCurve = regress_weight_curve(Hour, WeightDetrended))
  D = join(D, ddply(M, .(Hive,Date), regress_weight_daily))
  D = join(D, ddply(M, .(Hive,Date), summarize, WeightMin=min(Weight), WeightMax=max(Weight)))
  WD <<- D # Hack to get D into weight_at_end
  D = ddply(D, .(Hive,Date), mutate, WeightMidnight2= weight_at_end(Hive, Date))
  D = mutate(D, WeightSpan = WeightMax - WeightMin, 
                WeightGain = WeightMidnight2 - WeightMidnight)
  list(M,D)
}

# Add treaments to WD from files with columns 'Hive' and 'Treatment'
bc_add_treatments = function(D, file_names) {
  
  f = function(file_name) {
    read.table(file=file_name, header=TRUE, sep="\t")
  }
  
  T = ldply(file_names, f)
  D = join(D, T)
  D$Treatment = factor(D$Treatment, levels = c("Control","Low","Medium","High", "Thymol", "Combined"))
  D
}

# Read interventions from files with columns 'Date' and 'InterventionType'
bc_read_interventions = function(file_name, date_format) {
  M = read.table(file=file_name, header=TRUE, sep="\t", stringsAsFactors=FALSE)
  M$Date = date(parse_date_time(M$Date, date_format))
  M
}

# Return a data frame pointing out those values in vector x, that lie outside sd_max standard deviations
find_outliers = function(x, sd_max = 3.4) {
  M = data.frame(Value=x)
  M$Std=NA
  M$IsOutlier=NA
  ix = !is.na(x)
  M$Std[ix] = standardize(x[ix])
  M$IsOutlier[ix] = abs(M$Std[ix]) > sd_max
  M$IsOutlier[is.na(M$IsOutlier)] = FALSE
  M
}

# Reorder levels
reorder_levels = function(the_factor, new_order) {
  factor(the_factor, levels(the_factor)[new_order])
}

# Extract breaks from segmented model
get_breaks = function(model) {
  model$psi[,2]
}
  
# Extract slopes from segmented model
get_slopes = function(model) {
  slope(model)[[1]][,1]
}

# Extracts segments from a segmented model;
# returns the number of segments and three data frames (or NULL):
# - breaks with SE
# - slopes with SE and intercepts
# - (x,y) data to plot each line
# Note that the segments do not connect even if gap is set false;
# this is corrected through a calculation of the displacement
extract_segments = function(model) {
  if (is.null(model)) return(NULL)
  breaks = get_breaks(model)
  slopes = get_slopes(model)
  # intercepts = intercept(model, gap=FALSE)[[1]][,1]
  intercepts = intercept(model)[[1]][,1]
  x = model$model$Hour
  pred = predict(model)
  endings = c(min(x), breaks, max(x))
  n = length(slopes)
  line_plot = data.frame()
  # Loop through segments
  for (i in 1:n) {
    a = slopes[i]
    b = intercepts[i]
    x1 = endings[i]
    x2 = endings[i+1]
    y1 = a*x1+b
    y2 = a*x2+b
    ix_pred = min(which(x>x1))
    displacement = a*x[ix_pred] + b - pred[ix_pred]
    line_plot = rbind(line_plot, data.frame(Line=paste0("L",i), X=x1, Y=y1-displacement))
    line_plot = rbind(line_plot, data.frame(Line=paste0("L",i), X=x2, Y=y2-displacement))
  }
  rownames(line_plot) = NULL
  list(
    NumSegments = n,
    Breaks = data.frame(Value=breaks, 
               SE=model$psi[,3],
               row.names=1:(n-1)),
    Lines = data.frame(Slope=slopes,
               SlopeSE=slope(model)[[1]][,2],
               Intercept=intercepts,
               row.names=1:n),
    LinePlot = line_plot
  )
}

# Find SLR model by hive and date
find_slr_model = function(hive, date_) {
  ix_hive = which(sapply(slr_models, function(x) x$Hive==hive))
  ix_date = which(sapply(slr_models, function(x) x$Date==date_))
  ix = intersect(ix_hive, ix_date)
  validate_that(length(ix) == 1, msg=paste("More than one SLR model for hive and date:", hive, date_))
  slr_models[[ix]]
}

final_hour = function(hive, date) {
  sunrise = get_sunrise(hive, date)
  if (length(sunrise) == 1) sunrise + until_hour_after_sunrise else NA
}
